# Generated by Django 3.2.19 on 2023-05-08 16:39

import autoslug.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0041_offer_billing_start_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='offer',
            name='is_promotional',
            field=models.BooleanField(default=False, help_text='You can mark this offer as promotional to help identify it between normal priced and discount priced offers.', verbose_name='Is Promotional'),
        ),
        migrations.AlterField(
            model_name='offer',
            name='slug',
            field=autoslug.fields.AutoSlugField(editable=True, populate_from='name', unique_with=('site__id',)),
        ),
    ]



# Generated by Django 3.1.3 on 2021-01-07 23:09

from django.db import migrations, models

import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0017_auto_20201203_2107'),
    ]

    operations = [
        migrations.AddField(
            model_name='Address',
            name='uuid',
            field=models.UUIDField(editable=True, unique=False, verbose_name='UUID', null=True),
        ),
        migrations.AddField(
            model_name='CustomerProfile',
            name='uuid',
            field=models.UUIDField(editable=True, unique=False, verbose_name='UUID', null=True),
        )
    ]



# Generated by Django 3.1 on 2020-09-30 20:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0004_offer_offer_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='address',
            name='first_name',
            field=models.CharField(blank=True, max_length=150, verbose_name='First Name'),
        ),
        migrations.AddField(
            model_name='address',
            name='last_name',
            field=models.CharField(blank=True, max_length=150, verbose_name='Last Name'),
        ),
    ]



# Generated by Django 3.1.14 on 2021-12-16 16:29

from django.db import migrations


def add_trial_day_key_offer_term_details(apps, schema_editor):
    OfferModel = apps.get_model('vendor', 'Offer')
    for offer in OfferModel.objects.all():
        if 'trial_days' not in offer.term_details:
            offer.term_details['trial_days'] = 0
            offer.save()


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0027_bugfix_vendor_notes_invoice_history'),
    ]

    operations = [
        migrations.RunPython(add_trial_day_key_offer_term_details, reverse_code=migrations.RunPython.noop),
    ]



# Generated by Django 3.2.13 on 2022-06-07 16:49

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0033_payment_status'),
    ]

    operations = [
        migrations.CreateModel(
            name='Subscription',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, verbose_name='date created')),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='last updated')),
                ('deleted', models.BooleanField(default=False, verbose_name='Deleted')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True, verbose_name='UUID')),
                ('gateway_id', models.CharField(max_length=80, verbose_name='Subscription Gateway ID')),
                ('auto_renew', models.BooleanField(default=False, verbose_name='Auto Renew')),
                ('meta', models.JSONField(blank=True, default=dict, null=True, verbose_name='Meta')),
                ('profile', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='subscriptions', to='vendor.customerprofile', verbose_name='Purchase Profile')),
                ('status', models.IntegerField(choices=[(10, 'Pause'), (20, 'Active'), (30, 'Canceled'), (40, 'Suspended')], default=0, verbose_name='Status')),
            ],
            options={
                'verbose_name': 'Subscription',
                'verbose_name_plural': 'Subscriptions',
            },
        ),
    ]



# Generated by Django 3.1.3 on 2020-11-12 01:38

from django.db import migrations, models
import django.db.models.deletion
import vendor.models.utils


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('vendor', '0009_auto_20201111_0743'),
    ]

    operations = [
        migrations.AlterField(
            model_name='customerprofile',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='customer_profile', to='sites.site', verbose_name='Site'),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='invoices', to='sites.site', verbose_name='Site'),
        ),
        migrations.AlterField(
            model_name='offer',
            name='site',
            field=models.ForeignKey(default=vendor.models.utils.set_default_site_id, on_delete=django.db.models.deletion.CASCADE, related_name='product_offers', to='sites.site', verbose_name='Site'),
        ),
        migrations.AlterField(
            model_name='payment',
            name='billing_address',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='vendor.address', verbose_name='Billing Address'),
        ),
        migrations.AlterField(
            model_name='payment',
            name='created',
            field=models.DateTimeField(auto_now_add=True, verbose_name='Date Created'),
        ),
        migrations.AlterField(
            model_name='price',
            name='cost',
            field=models.FloatField(blank=True, null=True, verbose_name='Cost'),
        ),
        migrations.AlterField(
            model_name='price',
            name='offer',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='prices', to='vendor.offer', verbose_name='Offer'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='meta',
            field=models.JSONField(default=dict, verbose_name='Meta'),
        ),
        migrations.AlterField(
            model_name='taxclassifier',
            name='taxable',
            field=models.BooleanField(verbose_name='Taxable'),
        ),
    ]



from django.utils.module_loading import import_string
from django.core.exceptions import ObjectDoesNotExist

from vendor.config import PaymentProcessorSiteConfig
from vendor.processors.authorizenet import AuthorizeNetProcessor
from vendor.processors.stripe import StripeProcessor, StripeQueryBuilder
from vendor.processors.base import PaymentProcessorBase
from vendor.processors.dummy import DummyProcessor
from siteconfigs.models import SiteConfigModel


def get_site_payment_processor(site):
    site_processor = PaymentProcessorSiteConfig(site)
    return import_string(
        f"vendor.processors.{site_processor.get_key_value('payment_processor')}"
    )



# Generated by Django 3.2.19 on 2023-05-09 11:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0042_offer_is_promotional'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='global_discount',
            field=models.FloatField(blank=True, default=0, null=True, verbose_name='Global Discount'),
        ),
    ]



import math
import uuid
from datetime import timedelta

from autoslug import AutoSlugField
from django.contrib.sites.managers import CurrentSiteManager
from django.contrib.sites.models import Site
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from vendor.config import DEFAULT_CURRENCY
from vendor.utils import get_future_date_days, get_future_date_months

from .base import CreateUpdateModelBase, SoftDeleteModelBase
from .choice import TermDetailUnits, TermType
from .modelmanagers import (ActiveCurrentSiteManager, ActiveManager,
                            CurrentSiteSoftDeleteManager)
from .utils import is_currency_available, set_default_site_id


#########
# OFFER
#########
def offer_term_details_default():
    """
    Sets the default term values as a monthly subscription for a
    period of 12 months, with 0 trial months
    """
    return {
        "period_length": 1,
        "payment_occurrences": 12,
        "term_units": TermDetailUnits.MONTH,
        "trial_occurrences": 0,
        "trial_amount": 0,
        "trial_days": 0,
    }


class Offer(SoftDeleteModelBase, CreateUpdateModelBase):
    '''
    Offer attaches to a record from the designated VENDOR_PRODUCT_MODEL.
    This is so more than one offer can be made per product, with different
    priorities.  It also allows for the bundling of several products into
    a single Offer on the site.
    '''
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True)                                # Used to track the product
    slug = AutoSlugField(populate_from='name', unique_with='site__id', editable=True)                                               # SEO friendly
    site = models.ForeignKey(Site, verbose_name=_("Site"), on_delete=models.CASCADE, default=set_default_site_id, related_name="product_offers")                      # For multi-site support
    name = models.CharField(_("Name"), max_length=80, blank=True)                                           # If there is only a Product and this is blank, the product's name will be used, oterhwise it will default to "Bundle: <product>, <product>""
    start_date = models.DateTimeField(_("Start Date"), help_text=_("What date should this offer become available?"))
    end_date = models.DateTimeField(_("End Date"), blank=True, null=True, help_text=_("Expiration Date?"))
    terms = models.IntegerField(_("Terms"), default=0, choices=TermType.choices)
    term_details = models.JSONField(_("Term Details"), default=offer_term_details_default, blank=True, null=True, help_text=_("term_units: 10/20(Day/Month), trial_occurrences: 1(defualt)"))
    term_start_date = models.DateTimeField(_("Term Start Date"), help_text=_("When is this product available to use?"), blank=True, null=True)  # Useful for Event Tickets or Pre-Orders
    billing_start_date = models.DateTimeField(_("Billing Start Date"), blank=True, null=True, default=None, help_text=_("The start date to begin billing"))
    available = models.BooleanField(_("Available"), default=False, help_text=_("Is this currently available?"))
    bundle = models.BooleanField(_("Is a Bundle?"), default=False, help_text=_("Is this a product bundle? (auto-generated)"))  # Auto-generated based on if the count of the products is greater than 1.
    offer_description = models.TextField(_("Offer Description"), default=None, blank=True, null=True, help_text=_("You can enter a list of descriptions. Note: if you inputs something here the product description will not show up."))
    list_bundle_items = models.BooleanField(_("List Bundled Items"), default=False, help_text=_("When showing to customers, display the included items in a list?"))
    allow_multiple = models.BooleanField(_("Allow Multiple Purchase"), default=False, help_text=_("Confirm the user wants to buy multiples of the product where typically there is just one purchased at a time."))
    is_promotional = models.BooleanField(_("Is Promotional"), default=False, help_text=_("You can mark this offer as promotional to help identify it between normal priced and discount priced offers."))
    meta = models.JSONField(_("Meta"), default=dict, blank=True, null=True)

    objects = models.Manager()
    on_site = CurrentSiteManager()
    active = ActiveManager()
    on_site_active = ActiveCurrentSiteManager()
    on_site_not_deleted = CurrentSiteSoftDeleteManager()

    class Meta:
        verbose_name = "Offer"
        verbose_name_plural = "Offers"

    def __str__(self):
        return self.name

    def get_status_display(self):
        if self.end_date is not None and timezone.now() > self.end_date:
            return _("Expired")
        elif timezone.now() < self.start_date:
            return _("Scheduled")
        elif timezone.now() >= self.start_date:
            return _("Active")

    def get_msrp(self, currency=DEFAULT_CURRENCY):
        """
        Gets the sum of the products msrp cost for products.
        It assumes that all product in a offer use the same currency
        """
        currency = self.get_best_currency(currency)

        if self.products.count():
            return sum([product.get_msrp(currency) for product in self.products.all()])
        else:
            return 0

    def current_price(self, currency=DEFAULT_CURRENCY):
        '''
        Finds the highest priority active price and returns that, otherwise returns msrp total.
        '''
        now = timezone.now()
        price = self.prices.filter(Q(start_date__lte=now) | Q(start_date=None),
                                   Q(end_date__gte=now) | Q(end_date=None),
                                   Q(currency=currency)).order_by('-priority').first()  # first()/last() returns the model object or None

        if price is None:
            # If there is no price for the offer, all MSRPs should be summed up for the "price".
            return self.get_msrp(currency)

        elif price.cost is None:
            # If there is no price for the offer, all MSRPs should be summed up for the "price".
            return self.get_msrp(currency)

        return price.cost

    def get_current_price_instance(self, currency=DEFAULT_CURRENCY):
        now = timezone.now()
        price = self.prices.filter(Q(start_date__lte=now) | Q(start_date=None),
                                   Q(end_date__gte=now) | Q(end_date=None),
                                   Q(currency=currency)).order_by('-priority').first()  # first()/last() returns the model object or None

        return price

    def add_to_cart_link(self):
        return reverse("vendor_api:add-to-cart", kwargs={"slug": self.slug})

    def remove_from_cart_link(self):
        return reverse("vendor_api:remove-from-cart", kwargs={"slug": self.slug})

    def set_name_if_empty(self):
        product_names = [ product.name for product in self.products.all() ]
        if len(product_names) == 1:
            self.name = product_names[0]
        else:
            self.name = "Bundle: " + ", ".join(product_names)

    @property
    def description(self):
        if self.offer_description:
            return self.offer_description
        elif self.products.count():
            return self.products.all().first().description.get("description", "")
        else:
            return ""

    def discount(self, currency=DEFAULT_CURRENCY):
        """
        Gets the savings between the difference between the product's msrp and the current price
        """
        if self.is_promotional and self.current_price() < 0:
            return math.fabs(self.current_price())
        elif self.is_promotional and self.current_price() > 0:
            return 0
        
        discount = self.get_msrp(currency) - self.current_price(currency)

        if discount <= 0:
            return 0
            
        return discount

    def get_best_currency(self, currency=DEFAULT_CURRENCY):
        """
        Gets best currency for products available in this offer
        """
        product_msrp_currencies = [ set(product.meta['msrp'].keys()) for product in self.products.all() ]

        if product_msrp_currencies and len(product_msrp_currencies[0]) >= 2:  # fixes IndexError: list index out of range
            if is_currency_available(product_msrp_currencies[0].union(*product_msrp_currencies[1:]), currency=currency):
                return currency

        return DEFAULT_CURRENCY

    def get_trial_amount(self):
        return self.term_details.get('trial_amount', 0)

    def get_trial_savings(self):
        """
        Returns the savings compared to the current_price of the offer.
        """
        if not self.has_trial_occurrences():
            return 0

        trial_savings = self.current_price() - self.term_details.get('trial_amount', 0)

        if trial_savings < 0:
            return 0

        return trial_savings

    def get_trial_discount(self):
        """
        Returns the trial_amount if the offer has any trial occurrences.
        """
        if not self.has_trial_occurrences():
            return 0

        return self.term_details.get('trial_amount', 0)

    def get_trial_duration_in_months(self):
        duration = self.term_details.get('trial_days', 0)

        if duration <= 0:
            return 0

        return math.ceil(duration/31)

    def has_trial_occurrences(self):
        if self.term_details.get('trial_occurrences', 0) > 0:
            return True
        return False

    def get_next_billing_date(self):
        start_date = timezone.now()

        if self.term_start_date and self.term_start_date > start_date:
            start_date = self.term_start_date

        return self.get_offer_end_date(start_date)

    def get_period_length(self):
        if self.terms == TermType.SUBSCRIPTION:
            return self.term_details['period_length']
        else:
            return self.terms - 100   # You subtract 100 because enum are numbered according to their period length. eg Month = 101 and Year = 112

    def get_payment_occurrences(self):
        """
        Gets the defined payment ocurrences for a Subscription. It defaults to
        9999 which means it will charge that amount until the customer cancels the subscription.
        """
        return self.term_details.get('payment_occurrences', 9999)

    def get_trial_occurrences(self):
        return self.term_details.get('trial_occurrences', 0)
    
    def get_trial_days(self):
        return self.term_details.get('trial_days', 0)

    def has_any_discount_or_trial(self):
        if self.discount() or self.get_trial_amount() or\
           self.get_trial_occurrences() or self.get_trial_days() or\
           (self.billing_start_date and self.billing_start_date > timezone.now()):
            return True
            
        return False
    
    def has_trial(self):
        if self.get_trial_occurrences() or self.get_trial_days():
            return True
        return False
    
    def has_valid_billing_start_date(self, today=timezone.now()):
        if self.billing_start_date and self.billing_start_date > today:
            return True
        return False
    
    def get_offer_start_date(self, start_date=timezone.now()):
        if self.term_start_date and self.term_start_date > start_date:
            return self.term_start_date
        
        return start_date

    def get_offer_end_date(self, start_date=timezone.now()):
        """
        Determines the start date offset so the payment gateway starts charging the monthly offer
        """
        units = self.term_details.get('term_units', TermDetailUnits.MONTH)
        
        if units == TermDetailUnits.MONTH:
            return get_future_date_months(start_date, self.get_period_length())
        
        elif units == TermDetailUnits.DAY:
            return get_future_date_days(start_date, self.get_period_length())
    
    def get_trial_end_date(self, start_date=timezone.now()):
        if self.billing_start_date and self.billing_start_date > start_date:
            return self.billing_start_date - timedelta(days=1)
        
        trial_occurrences = self.get_trial_occurrences()
        if trial_occurrences:
            units = self.term_details.get('term_units', TermDetailUnits.MONTH)

            if units == TermDetailUnits.MONTH:
                return get_future_date_months(start_date, self.get_period_length() * trial_occurrences)

            elif units == TermDetailUnits.DAY:
                return get_future_date_days(start_date, self.get_period_length() * trial_occurrences)

        return start_date + timedelta(days=self.get_trial_days())

    def get_payment_start_date_trial_offset(self, start_date=timezone.now()):
        trial_end_date = self.get_trial_end_date(start_date)

        if trial_end_date == start_date:
            return start_date
        
        return trial_end_date + timedelta(days=1)



# Generated by Django 3.1.3 on 2021-09-14 12:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vendor', '0025_auto_20210914_0025'),
    ]

    operations = [
        migrations.AddField(
            model_name='invoice',
            name='deleted',
            field=models.BooleanField(default=False, verbose_name='Deleted'),
        ),
        migrations.AddField(
            model_name='payment',
            name='deleted',
            field=models.BooleanField(default=False, verbose_name='Deleted'),
        ),
        migrations.AddField(
            model_name='receipt',
            name='deleted',
            field=models.BooleanField(default=False, verbose_name='Deleted'),
        ),
    ]



from django.apps import AppConfig
class VendorConfig(AppConfig):
    name = 'vendor'

    def ready(self):
        from vendor.config import VENDOR_PAYMENT_PROCESSOR, SupportedPaymentProcessor

        if VENDOR_PAYMENT_PROCESSOR == SupportedPaymentProcessor.STRIPE:
            import vendor.signals.stripe_signals
    
    #  TODO: Would be better to find way to check if any sites have stripe configured instead of having to set the VENDOR_PAYMENT_PROCESSOR to stripe 
    # to enable it's signals. You can't do the following way, because you assume that you have already ran migration, but if you are in a fresh install
    # this code will error out any time you run a manage.py command because Site has not been migrated and you are trying to import it. 
    # def ready(self):
    #     from django.contrib.sites.models import Site
    #     from vendor.config import SupportedPaymentProcessor
    #     from vendor.processors import get_site_payment_processor

    #     is_stripe_configured = next((True for site in Site.objects.all() if get_site_payment_processor(site) == SupportedPaymentProcessor.STRIPE), None)

    #     if is_stripe_configured:
    #         import vendor.signals.stripe_signals


import json
import logging
import stripe

from math import modf
from django.conf import settings
from django.utils import timezone
from vendor.config import DEFAULT_CURRENCY, StripeConnectAccountConfig, VendorSiteCommissionConfig
from vendor.integrations import StripeIntegration
from vendor.models import Offer, CustomerProfile
from vendor.models.choice import (
    Country,
    TermType,
    TermDetailUnits,
)
from vendor.processors.base import PaymentProcessorBase

logger = logging.getLogger(__name__)


class StripeQueryBuilder:
    """
    Query builder that adheres to Stripe search rules found here https://stripe.com/docs/search

    Ex:
    To generate a query like 'name:"Johns Offer" AND metadata["site"]:"site4"'

    query_builder = StripeQueryBuilder()

    name_clause = query_builder.make_clause_template()
    name_clause['field'] = 'name'
    name_clause['value'] = 'Johns Offer'
    name_clause['operator'] = query_builder.EXACT_MATCH
    name_clause['next_operator'] = query_builder.AND

    metadata_clause = query_builder.make_clause_template()
    metadata_clause['field'] = 'metadata'
    metadata_clause['key'] = 'site'
    metadata_clause['value'] = 'site4'
    metadata_clause['operator'] = query_builder.EXACT_MATCH

    query = query_builder.build_search_query(processor.stripe.Product, [name_clause, metadata_clause])


    """

    # Search syntax
    EXACT_MATCH = ':'
    AND = 'AND'
    OR = 'OR'
    EXCLUDE = '-'
    NULL = 'NULL'
    SUBSTRING_MATCH = '~'
    GREATER_THAN = '>'
    LESS_THAN = '<'
    EQUALS = '='
    GREATER_THAN_OR_EQUAL_TO = '>='
    LESS_THAN_OR_EQUAL_TO = '<='

    # Valid fields
    VALID_FIELDS = {
        'charge': [
            'amount',
            'billing_details.address.postal_code',
            'created',
            'currency',
            'customer',
            'disputed',
            'metadata',
            'payment_method_details.card.last4',
            'payment_method_details.card.exp_month',
            'payment_method_details.card.exp_year',
            'payment_method_details.card.brand',
            'payment_method_details.card.fingerprint',
            'refunded',
            'status'
        ],
        'customer': [
            'created',
            'email',
            'metadata',
            'name',
            'phone'
        ],
        'invoice': [
            'created',
            'currency',
            'customer',
            'metadata',
            'number',
            'receipt_number',
            'subscription',
            'total'
        ],
        'paymentintent': [
            'amount',
            'created',
            'currency',
            'customer',
            'metadata',
            'status',
        ],
        'price': [
            'active',
            'lookup_key',
            'currency',
            'product',
            'metadata',
            'type',
        ],
        'product': [
            'active',
            'description',
            'name',
            'shippable',
            'metadata',
            'url',
        ],
        'subscription': [
            'created',
            'metadata',
            'status',
        ],
    }

    def make_clause_template(self, field=None, operator=None, key=None, value=None, next_operator=None):
        return {
            'field': field,
            'operator': operator,
            'key': key,
            'value': value,
            'next_operator': next_operator
        }

    def is_valid_field(self, stripe_object_class, field):
        return field in self.VALID_FIELDS[stripe_object_class.__name__.lower()]

    def search_clause_checks_pass(self, clause_obj):
        """
        All checks should be added here to make sure the caller isnt missing required params
        """
        if clause_obj.get('field', None) == 'metadata':
            if not clause_obj.get('key', None):
                logger.error(f'StripeQueryBuilder.search_clause_checks_pass: metadata searches need a key field')
                return False

        # TODO add more checks

        return True

    def build_search_query(self, stripe_object_class, search_clauses):
        query = ""

        if not isinstance(search_clauses, list):
            logger.error(f'Passed in params {search_clauses} is not a list of search clauses (dicts)')
            return query

        if not len(search_clauses) > 0:
            logger.error(f'Passed in search clauses {search_clauses} cannot be empty')
            return query

        for query_obj in search_clauses:
            field = query_obj.get('field', None)
            operator = query_obj.get('operator', None)
            key = query_obj.get('key', None)
            value = query_obj.get('value', None)
            next_operator = query_obj.get('next_operator', None)

            if not self.search_clause_checks_pass(query_obj):
                logger.error(f'StripeQueryBuilder.build_search_query: search clause {query_obj} is not valid')
                return query

            if self.is_valid_field(stripe_object_class, field):
                if not key:
                    # not metadata
                    if isinstance(value, str):
                        query += f'{field}{operator}"{value}"'
                    else:
                        query += f'{field}{operator}{value}'
                else:
                    # is metadata
                    if isinstance(value, str):
                        query += f'{field}["{key}"]{operator}"{value}"'
                    else:
                        query += f'{field}["{key}"]{operator}{value}'

                if next_operator:
                    # space, AND, OR
                    query += f' {next_operator} '

            else:
                logger.error(f'StripeQueryBuilder.build_search_query: {field} is not valid for {stripe_object_class}')
                return query

        return query

class StripeProcessor(PaymentProcessorBase):
    """ 
    Implementation of Stripe SDK
    https://self.stripe.com/docs/api/authentication?lang=python
    """

    TRANSACTION_SUCCESS_MESSAGE = 'message'
    TRANSACTION_SUCCESS_CODE = 'code'
    TRANSACTION_FAIL_MESSAGE = 'error_text'
    TRANSACTION_FAIL_CODE = 'error_code'
    TRANSACTION_RESPONSE_CODE = 'response_code'
    source = None
    products_mapping = {}

    def processor_setup(self, site, source=None):
        self.credentials = StripeIntegration(site)
        self.source = source
        self.site = site
        self.stripe = stripe
        self.query_builder = StripeQueryBuilder()

        if self.credentials.instance:
            self.stripe.api_key = self.credentials.instance.private_key
        elif settings.STRIPE_SECRET_KEY:
            self.stripe.api_key = settings.STRIPE_SECRET_KEY
        else:
            logger.error("StripeProcessor missing keys in settings: STRIPE_PUBLIC_KEY")
            raise ValueError("StripeProcessor missing keys in settings: STRIPE_PUBLIC_KEY")

        # TODO handle this better, but needed to get passed this error message
        # Search is not supported on api version 2016-07-06. Update your API version, or set the API Version of this request to 2020-08-27 or greater.
        # https://stripe.com/docs/libraries/set-version
        self.stripe.api_version = '2022-08-01'


    def customer_setup(self):
        if not self.invoice.profile.meta.get('stripe_id'):
            self.create_stripe_customers([self.invoice.profile])
        
        stripe_customer = self.stripe_get_object(self.stripe.Customer, self.invoice.profile.meta.get('stripe_id'))

        if not stripe_customer:
            self.create_stripe_customers([self.invoice.profile])

    def subscription_offer_setup(self):
        offers_to_sync = []
        for order_item in self.invoice.order_items.all():
            if not order_item.offer.meta.get('stripe'):
                offers_to_sync.append(order_item.offer)

        self.create_offers(offers_to_sync)

    ##########
    # Parsers
    ##########
    def parse_response(self, response_data):
        """
        Processes the transaction response
        """
        ...


    ##########
    # Stripe utils
    ##########
    def stripe_call(self, *args):
        func, func_args = args
        self.transaction_succeeded = False

        try:
            if isinstance(func_args, str):
                self.transaction_response = func(func_args)
            else:
                self.transaction_response = func(**func_args)

        except (self.stripe.error.CardError, self.stripe.error.RateLimitError,
                self.stripe.error.InvalidRequestError, self.stripe.error.AuthenticationError,
                self.stripe.error.APIConnectionError, self.stripe.error.StripeError,
                Exception) as e:

            user_message = ""
            if hasattr(e, 'user_message'):
                user_message = e.user_message
            raw = {
                'exception': f"{e}\n{func}\n{func_args}:user_message: {user_message}"
            }
            errors = {
                'error': f"{e}"
            }
            self.transaction_info = self.get_transaction_info(raw=raw, errors=errors)
            logger.error(self.transaction_info)
            return None

        self.transaction_succeeded = True
        self.transaction_info = self.get_transaction_info(raw=f"{func} - {func_args} {json.dumps(self.transaction_response)}", data=self.transaction_response.data if 'data' in self.transaction_response else "")

        return self.transaction_response

    def convert_decimal_to_integer(self, decimal):
        if decimal == 0:
            return 0
            
        fraction_part, whole_part = modf(decimal)
                
        whole_str = str(whole_part).split('.')[0]
        fraction_str = str(fraction_part).split('.')[1][:2]

        if len(fraction_str) < 2:
            fraction_str = "0" + fraction_str

        stripe_amount = int("".join([whole_str, fraction_str]))
            
        return stripe_amount

    def get_stripe_connect_account(self):
        stripe_connect = StripeConnectAccountConfig(self.site)

        if stripe_connect.instance:
            return stripe_connect.get_key_value('stripe_connect_account')

        return None

    def get_application_fee_percent(self):
        vendor_site_commission = VendorSiteCommissionConfig(self.site)

        if vendor_site_commission.instance:
            return vendor_site_commission.get_key_value('commission')

        return None

    def get_application_fee_amount(self):
        vendor_site_commission = VendorSiteCommissionConfig(self.site)

        if vendor_site_commission.instance:
            return self.convert_decimal_to_integer((vendor_site_commission.get_key_value('commission')*self.invoice.total)/100)

        return None

    ##########
    # CRUD Stripe Object
    ##########
    def stripe_create_object(self, stripe_object_class, object_data):
        stripe_object = self.stripe_call(stripe_object_class.create, object_data)

        return stripe_object

    def stripe_query_object(self, stripe_object_class, query, limit=100, page=None):
        if isinstance(query, dict):
            query_data = query
            query_data['limit'] = limit
            query_data['page'] = page

        elif isinstance(query, str):
            query_data = {
                'query': query,
                'limit': limit,
                'page': page
            }

        else:
            logger.error(f'stripe_query_object: {query} is invalid')
            return None

        query_result = self.stripe_call(stripe_object_class.search, query_data)

        return query_result

    def stripe_delete_object(self, stripe_object_class, object_id):
        delete_result = self.stripe_call(stripe_object_class.delete, object_id)

        return delete_result

    def stripe_get_object(self, stripe_object_class, object_id, expand=None):
        if expand:
            # some fields are expandable https://stripe.com/docs/api/expanding_objects
            object_id = {
                'id': object_id,
                'expand': expand
            }
        stripe_object = self.stripe_call(stripe_object_class.retrieve, object_id)

        return stripe_object

    def stripe_update_object(self, stripe_object_class, object_id, object_data):
        object_data['sid'] = object_id
        stripe_object = self.stripe_call(stripe_object_class.modify, object_data)

        if self.transaction_succeeded:
            self.transaction_succeeded = False
            return stripe_object

        return None

    def stripe_list_objects(self, stripe_object_class, limit=100, starting_after=None):
        object_data = {
            'limit': limit,
            'starting_after': starting_after
        }
        stripe_objects = self.stripe_call(stripe_object_class.list, object_data)

        return stripe_objects

    def get_all_stripe_list_objects(self, stripe_object):
        """
        Get entire list of any stripe object with .list() method.
        Will make multiple stripe calls until the list is exhausted
        """
        object_list = []
        starting_after = None
        while True:
            objs = self.stripe_list_objects(stripe_object, limit=100, starting_after=starting_after)
            object_list.extend(objs)
            if objs['has_more']:
                starting_after = objs['data'][-1]['id']
            else:
                break

        return object_list

    def get_all_stripe_search_objects(self, stripe_object, query, limit=100):
        """
        Get entire list of any stripe object with .search() method.
        Will make multiple stripe calls until the list is exhausted
        """
        object_list = []
        page = None

        while True:
            objs = self.stripe_query_object(stripe_object, query=query, limit=limit, page=page)
            object_list.extend(objs['data'])
            if objs['has_more']:
                page = objs['next_page']
            else:
                break

        return object_list
        
    # Stripe Object Builders
    ##########
    def build_transfer_data(self):
        return {
            'destination': self.get_stripe_connect_account(),
            # Not required if you are using the application_fee parameter
            # 'amount_percent': 100 - self.get_application_fee_percent()
        }

    def build_customer(self, customer_profile):
        return {
            'name': f"{customer_profile.user.first_name} {customer_profile.user.last_name}",
            'email': customer_profile.user.email,
            'metadata': {'site': customer_profile.site}
        }
    
    def build_product(self, offer):
        return {
            'name': offer.name,
            'metadata': {'site': offer.site.domain, 'pk': offer.pk}

        }
    # TODO: Try to reduce the number of arguments to max 3. 
    def build_price(self, offer, mspr, current_price, currency, price_pk=None):
        if 'stripe' not in offer.meta or 'product_id' not in offer.meta['stripe']:
            raise TypeError(f"Price cannot be created without a product_id on offer.meta['stripe'] field")

        price_data = {
            'active': True,
            'product': offer.meta['stripe']['product_id'],
            'currency': currency,
            'unit_amount': self.convert_decimal_to_integer(current_price),
            'metadata': {
                'site': offer.site.domain,
                'msrp': mspr
                }
        }

        if price_pk:
            price_data['metadata']['pk'] = price_pk
        
        if offer.terms < TermType.PERPETUAL:
            price_data['recurring'] = {
                'interval': 'month' if offer.term_details['term_units'] == TermDetailUnits.MONTH else 'year',
                'interval_count': offer.term_details['period_length'],
                'usage_type': 'licensed'
            }
        
        return price_data
    
    def is_card_valid(self):
        # TODO: see how stripe has check for 
        return True

    def build_coupon(self, offer, currency):
        coupon_data = {
            'name': offer.name,
            'currency': currency,
            'amount_off': self.convert_decimal_to_integer(offer.discount(currency)),
            'metadata': {'site': offer.site.domain}
        }

        if offer.terms < TermType.PERPETUAL:
            coupon_data['amount_off'] = self.convert_decimal_to_integer(offer.get_trial_discount())
            coupon_data['duration'] = 'once' if offer.term_details['trial_occurrences'] <= 1 else 'repeating'
            coupon_data['duration_in_months'] = offer.term_details['trial_occurrences']
        
        return coupon_data

    def build_payment_method(self):
        return {
            'type': 'card',
            'card': {
                'number': self.payment_info.cleaned_data.get('card_number'),
                'exp_month': self.payment_info.cleaned_data.get('expire_month'),
                'exp_year': self.payment_info.cleaned_data.get('expire_year'),
                'cvc': self.payment_info.cleaned_data.get('cvv_number'),
            },
            'billing_details': {
                'address': {
                    'line1': self.billing_address.cleaned_data.get('address_1', None),
                    'line2': self.billing_address.cleaned_data.get('address_2', None),
                    'city': self.billing_address.cleaned_data.get("locality", ""),
                    'state': self.billing_address.cleaned_data.get("state", ""),
                    'country': Country.names[Country.values.index(int(self.billing_address.cleaned_data.get("country")))],
                    'postal_code': self.billing_address.cleaned_data.get("postal_code")
                },
                'name': self.payment_info.cleaned_data.get('full_name', None),
                'email': self.invoice.profile.user.email
            }
        }

    def build_payment_intent(self, amount, currency=DEFAULT_CURRENCY):
        return {
            'amount': amount,
            'currency': currency,
            'customer': self.invoice.profile.meta['stripe_id'],
            'application_fee_percent': self.get_application_fee_amount(),
            'transfer_data': self.get_stripe_connect_account(),
            'on_behalf_of': self.get_stripe_connect_account()
        }

    def build_setup_intent(self, payment_method_id):
        return {
            'customer': self.invoice.profile.meta['stripe_id'],
            'confirm': True,
            'payment_method_types': ['card'],
            'payment_method': payment_method_id,
            'metadata': {'site': self.invoice.site},
            'on_behalf_of': self.get_stripe_connect_account()
        }
    
    def build_subscription(self, subscription, payment_method_id):
        return {
            'customer': self.invoice.profile.meta['stripe_id'],
            'promotion_code': self.invoice.coupon_code.first().meta['stripe_id'] if self.invoice.coupon_code.count() else None,
            'items': [{'price': subscription.offer.meta['stripe']['price_id']}],
            'default_payment_method': payment_method_id,
            'metadata': {'site': self.invoice.site},
            'trial_period_days': subscription.offer.get_trial_days(),
            'application_fee_percent': self.get_application_fee_percent(),
            'transfer_data': self.build_transfer_data(),
            'on_behalf_of': self.get_stripe_connect_account()
        }

    def build_invoice_line_item(self, order_item, invoice_id):
        line_item = {
            'customer': self.invoice.profile.meta.get('stripe_id'),
            'invoice': invoice_id,
            'price': order_item.offer.meta['stripe'].get('price_id'),
        }

        if order_item.offer.has_trial() or order_item.offer.has_valid_billing_start_date() or order_item.offer.discount():
            line_item['discounts'] = [{'coupon': order_item.offer.meta['stripe'].get('coupon_id')}]

        return line_item

    def build_invoice(self, currency=DEFAULT_CURRENCY):
        return {
            'customer': self.invoice.profile.meta.get('stripe_id'),
            'currency': currency,
        }

    ##########
    # Customers
    ##########
    def get_stripe_customers(self, site, expand=None):
        """
        Returns all Stripe created customers for this site
        """

        clause = self.query_builder.make_clause_template(
            field='metadata',
            key='site',
            value=site.domain,
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Customer, [clause])

        customer_search = self.get_all_stripe_search_objects(self.stripe.Customer, {'query': query})

        return customer_search

        return []

    def get_vendor_customers_in_stripe(self, customer_email_list, site):
        """
        Returns all vendor customers who have been created as Stripe customers
        """
        users = CustomerProfile.objects.filter(
            user__email__iregex=r'(' + '|'.join(customer_email_list) + ')', # iregex used for case insensitive list match
            site=site
        ).select_related('user')

        return users

    def get_vendor_customers_not_in_stripe(self, customer_email_list, site):
        """
        Returns all vendor customers who have not been created as Stripe customers
        """
        users = CustomerProfile.objects.filter(
            site=site
        ).exclude(
            user__email__iregex=r'(' + '|'.join(customer_email_list) + ')',  # iregex used for case insensitive list match
        ).select_related('user')

        return users

    def create_stripe_customers(self, customers):
        for profile in customers:
            profile_data = self.build_customer(profile)
            new_stripe_customer = self.stripe_create_object(self.stripe.Customer, profile_data)

            if new_stripe_customer:
                profile_meta = profile.meta
                profile_meta['stripe_id'] = new_stripe_customer['id']
                CustomerProfile.objects.filter(pk=profile.pk).update(meta=profile_meta)

    def update_stripe_customers(self, customers):
        for profile in customers:
            customer_id = profile.meta.get('stripe_id')
            profile_data = self.build_customer(profile)
            existing_stripe_customer = self.stripe_update_object(self.stripe.Customer, customer_id, profile_data)

            if existing_stripe_customer:
                profile_meta = profile.meta
                profile_meta['stripe_id'] = existing_stripe_customer['id']
                CustomerProfile.objects.filter(pk=profile.pk).update(meta=profile_meta)
            else:
                self.create_stripe_customers([profile])
    
    def get_customer_payment_methods(self, customer_id):
        payment_methods = self.stripe_call(self.stripe.Customer.list_payment_methods, customer_id)

        if not payment_methods.data:
            return None

        return payment_methods.data
    
    ##########
    # Offers/Products
    ##########
    def does_product_exist(self, name, metadata):
        name_clause = self.query_builder.make_clause_template(
            field='name',
            value=name,
            operator=self.query_builder.EXACT_MATCH,
            next_operator=self.query_builder.AND
        )

        metadata_clause = self.query_builder.make_clause_template(
            field='metadata',
            key=metadata['key'],
            value=metadata['value'],
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Product, [name_clause, metadata_clause])

        search_data = self.stripe_query_object(self.stripe.Product, {'query': query})

        if search_data:
            if search_data.get('data', False):
                return True

        return False

    def get_site_offers(self, site):
        """
        Returns all Stripe created Products
        """

        clause = self.query_builder.make_clause_template(
            field='metadata',
            key='site',
            value=site.domain,
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Product, [clause])

        product_search = self.get_all_stripe_search_objects(self.stripe.Product, {'query': query})

        return product_search

    def get_vendor_offers_in_stripe(self, offer_pk_list, site):
        offers = Offer.objects.filter(site=site, pk__in=offer_pk_list)
        return offers

    def get_vendor_offers_not_in_stripe(self, offer_pk_list, site):
        offers = Offer.objects.filter(site=site).exclude(pk__in=offer_pk_list)
        return offers

    def create_offers(self, offers):
        for offer in offers:
            offer_meta = offer.meta
            # build product first, since product_id is needed to build price later
            product_data = self.build_product(offer)

            new_stripe_product = self.stripe_create_object(self.stripe.Product, product_data)

            if new_stripe_product:
                if offer_meta.get('stripe'):
                    offer_meta['stripe']['product_id'] = new_stripe_product['id']
                else:
                    offer_meta['stripe'] = {'product_id': new_stripe_product['id']}

            # TODO: Need to explore what is the best way to upload prices in each currency
            # currently we will only support the default currency (DEFAULT_CURRENCY) se
            # on the vendor.config.py file
            # for currency in AVAILABLE_CURRENCIES:
            #     ...
            price = offer.get_current_price_instance() if offer.get_current_price_instance() else None
            msrp = offer.get_msrp()
            current_price = msrp
            price_pk = None
            
            if price:
                if not msrp and price.cost > msrp:
                    current_price = price.cost
                price_pk = price.pk

            price_data = self.build_price(offer, msrp, current_price, DEFAULT_CURRENCY, price_pk)
            new_stripe_price = self.stripe_create_object(self.stripe.Price, price_data)
            
            if new_stripe_price:
                if offer_meta.get('stripe'):
                    offer_meta['stripe']['price_id'] = new_stripe_price['id']
                else:
                    offer_meta['stripe'] = {'price_id': new_stripe_price['id']}

            if offer.discount() or offer.get_trial_amount():
                coupon_data = self.build_coupon(offer, DEFAULT_CURRENCY)

                new_stripe_coupon = self.stripe_create_object(self.stripe.Coupon, coupon_data)

                if new_stripe_coupon:
                    if offer_meta.get('stripe'):
                        offer_meta['stripe']['coupon_id'] = new_stripe_coupon['id']
                    else:
                        offer_meta['stripe'] = {'coupon_id': new_stripe_coupon['id']}
            
            Offer.objects.filter(pk=offer.pk).update(meta=offer_meta)  # Doing an update to avoid triggering post_save()

    def update_offers(self, offers):
        coupons = self.get_coupons()

        for offer in offers:
            # Handle product
            offer_meta = offer.meta
            product_id = offer.meta['stripe']['product_id']
            product_data = self.build_product(offer)
            existing_stripe_product = self.stripe_update_object(self.stripe.Product, product_id, product_data)
            
            if not existing_stripe_product:
                self.create_offers([offer])
            else:
                if offer_meta.get('stripe'):
                    offer_meta['stripe']['product_id'] = existing_stripe_product['id']
                else:
                    offer_meta['stripe'] = {'product_id': existing_stripe_product['id']}

                # Handle Price
                # TODO: Need to explore what is the best way to upload prices in each currency
                # currently we will only support the default currency (DEFAULT_CURRENCY) se
                # on the vendor.config.py file
                # for currency in AVAILABLE_CURRENCIES:
                #     ...
                msrp = offer.get_msrp()
                stripe_product_prices = self.get_stripe_prices_for_product(product_id)

                price = offer.get_current_price_instance() if offer.get_current_price_instance() else None
                current_price = msrp
                price_pk = None
                
                if price:
                    current_price = price.cost
                    price_pk = price.pk
                
                price_data = self.build_price(offer, msrp, current_price, DEFAULT_CURRENCY, price_pk)
                stripe_price = None

                if not stripe_product_prices:
                    stripe_price = self.stripe_create_object(self.stripe.Price, price_data)

                else:
                    prices_to_deactivate = [stripe_price for stripe_price in stripe_product_prices if self.convert_decimal_to_integer(current_price) != stripe_price.unit_amount]
                    prices_to_check = [stripe_price for stripe_price in stripe_product_prices if self.convert_decimal_to_integer(current_price) == stripe_price.unit_amount]
                    
                    for stripe_price_deactivate in prices_to_deactivate:
                        self.stripe_update_object(self.stripe.Price, stripe_price_deactivate.id, {'active': False})
                    
                    any_match = False
                    for stripe_price_check in prices_to_check:
                        if self.does_stripe_and_vendor_price_match(stripe_price_check, price_data):
                            any_match = True
                            stripe_price = stripe_price_check
                        else:
                            self.stripe_update_object(self.stripe.Price, stripe_price_check.id, {'active': False})
                    
                    if not any_match:
                        stripe_price = self.stripe_create_object(self.stripe.Price, price_data)

                if stripe_price:
                    if offer_meta.get('stripe'):
                        offer_meta['stripe']['price_id'] = stripe_price['id']
                    else:
                        offer_meta['stripe'] = {'price_id': stripe_price['id']}

                # Handle Coupon
                coupon_data = self.build_coupon(offer, DEFAULT_CURRENCY)
                discount = self.convert_decimal_to_integer(offer.discount())
                trial_days = offer.get_trial_days()

                # If this offer has a discount check if its on stripe to create, update, delete
                if discount or trial_days:
                    stripe_coupon_matches = self.match_coupons(coupons, offer)
                    if not stripe_coupon_matches:
                        stripe_coupon = self.stripe_create_object(self.stripe.Coupon, coupon_data)
                    elif len(stripe_coupon_matches) == 1:
                        coupon_id = stripe_coupon_matches[0]['id']
                        stripe_coupon = self.stripe_update_object(self.stripe.Coupon, coupon_id, coupon_data)
                    else:
                        # Duplicates, so delete all but one and update it
                        for coupon_data in stripe_coupon_matches[1:]:
                            self.stripe_delete_object(self.stripe.Coupon, coupon_data['id'])

                        # update the only one we have remaining
                        stripe_coupon = self.stripe_update_object(self.stripe.Coupon, stripe_coupon_matches[0]['id'], coupon_data)

                    if stripe_coupon:
                        if offer_meta.get('stripe'):
                            offer_meta['stripe']['coupon_id'] = stripe_coupon['id']
                        else:
                            offer_meta['stripe'] = {'coupon_id': stripe_coupon['id']}

                Offer.objects.filter(pk=offer.pk).update(meta=offer_meta)

    def get_product_id_with_name(self, name, metadata):
        name_clause = self.query_builder.make_clause_template(
            field='name',
            value=name,
            operator=self.query_builder.EXACT_MATCH,
            next_operator=self.query_builder.AND
        )
        metadata_clause = self.query_builder.make_clause_template(
            field='metadata',
            key=metadata['key'],
            value=metadata['value'],
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Product, [name_clause, metadata_clause])
        search_data = self.stripe_query_object(self.stripe.Product, {'query': query})

        if search_data:
            return search_data['data'][0]['id']
        return None

    ##########
    # Prices
    ##########
    def does_stripe_and_vendor_price_match(self, stripe_price, vendor_price):
        if not stripe_price.unit_amount == vendor_price['unit_amount']:
            return False
        
        if 'recurring' in vendor_price:
            if 'recurring' not in stripe_price:
                return False

            if not stripe_price.recurring['interval'] == vendor_price['recurring']['interval']:
                return False
            
            if not stripe_price.recurring['interval_count'] == vendor_price['recurring']['interval_count']:
                return False
        
        return True

    def get_price_id_with_product(self, product):
        price = self.stripe_get_object(self.stripe.Price, {'id': product})

        if price:
            return price['id']

        return None
    
    def get_stripe_prices_for_product(self, product):
        product_clause = self.query_builder.make_clause_template(
            field='product',
            value=product,
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Price, [product_clause])

        search_data = self.get_all_stripe_search_objects(self.stripe.Price, {'query': query})

        return search_data

    def get_customer_email(self, customer_id):
        customer = self.stripe_get_object(self.stripe.Customer, customer_id)

        if customer:
            return customer['email']
        return None

    def get_customer_id_for_expiring_cards(self, month):
        # month is passed as YYYY-M

        customers = self.get_stripe_customers(self.site)
        sources = [
            (customer['id'], customer['email'], customer['default_source'])
            for customer in customers if customer.get('default_source')
        ]

        expired_cards_emails = []
        for customer_id, customer_email, source_id in sources:
            if 'card' in source_id:
                stripe_card = self.stripe_call(
                    self.stripe.Customer.retrieve_source,
                    {'id': customer_id, 'nested_id': source_id}
                )
                if stripe_card:
                    if stripe_card['exp_year'] and stripe_card['exp_month']:
                        exp_date = f"{stripe_card['exp_year']}-{stripe_card['exp_month']}"
                        exp_date_obj = timezone.datetime.strptime(exp_date, "%Y-%m")
                        current_date_obj = timezone.datetime.strptime(month, "%Y-%m")
                        date_diff = (exp_date_obj - current_date_obj).days
                        if 0 < date_diff <= 60:
                            expired_cards_emails.append(customer_email)

        return expired_cards_emails

    def does_price_exist(self, product, metadata):
        product_clause = self.query_builder.make_clause_template(
            field='product',
            value=product,
            operator=self.query_builder.EXACT_MATCH,
            next_operator=self.query_builder.AND
        )
        metadata_clause = self.query_builder.make_clause_template(
            field='metadata',
            key=metadata['key'],
            value=metadata['value'],
            operator=self.query_builder.EXACT_MATCH
        )

        query = self.query_builder.build_search_query(self.stripe.Price, [product_clause, metadata_clause])

        search_data = self.stripe_query_object(self.stripe.Price, {'query': query})

        if search_data:
            if search_data.get('data', False):
                return True

        return False

    def create_price_with_product(self, product):
        price_data = {
            'currency': self.invoice.currency,
            'product': product
        }

        price = self.stripe_create_object(self.stripe.Price, price_data)

        if price:
            return price['id']

        return None

    ##########
    # Coupons
    ##########
    def match_coupons(self, coupons, offer):
        matches = []
        amount_off = self.convert_decimal_to_integer(offer.discount())
        duration = offer.get_trial_duration_in_months()
        for coupon in coupons:
            if coupon['metadata']['site'] == offer.site.domain and coupon['amount_off'] == amount_off\
                    and coupon['duration_in_months'] == duration:
                matches.append(coupon)
        return matches

    def get_coupons(self):
        """
        Returns all stripe Coupons

        This is needed since there is no search method on Coupon.
        """

        return self.get_all_stripe_list_objects(self.stripe.Coupon) or []

    ##########
    # Sync Vendor and Stripe
    ##########
    def initialize_products(self, site):
        """
        Grab all subscription offers on invoice and either create or fetch Stripe products.
        Then using those products, create Stripe prices. Add all that to products_mapping

        Will be used in create_subscription
        """
        # Create Customer loop through all site customer and save their id's
        # Create Stripe product with site offers
        ## Create Stripe Price from Offer.Prices
        ## Create Coupons from Offer.term_details. 
        for subscription in self.invoice.get_recurring_order_items():
            product_name = subscription.offer.name
            product_name_full = f'{product_name} - site {self.site.domain}'
            product_details = subscription.offer.term_details
            total = self.to_valid_decimal(subscription.total - subscription.discounts)
            interval = "month" if product_details['term_units'] == TermDetailUnits.MONTH else "year"
            metadata = {
                'key': 'site',
                'value': 'site4',
                'field': 'metadata'
            }
            product_response = self.does_product_exist(product_name_full, metadata=metadata)
            if product_response:
                product_id = self.get_product_id_with_name(product_name_full)

                # Each product needs an attached price obj. Check if one exists for the product
                # and attach it to the mapping or create one and attach
                if product_id:
                    price_response = self.does_price_exist(product_id)
                    if price_response:
                        price_id = self.get_price_id_with_product(product_id)
                    else:
                        price_id = self.create_price_with_product(product_id)

                    self.products_mapping[product_name] = {
                        'product_id': product_id,
                        'price_id': price_id
                    }

            else:
                # product doesnt exist, create it and related pricing obj
                product = self.stripe_call(self.stripe.Product.create, {
                    'name': product_name,
                    'metadata': product_details,
                    'default_price_data': {
                        'currency': self.invoice.currency,
                        'unit_amount_decimal': total,
                        'recurring': {
                            'interval': interval,
                            'interval_count': subscription.offer.get_period_length()
                        }
                    }
                })
                if product:
                    product_id = product['id']

                    # Each product needs an attached price obj. Check if one exists for the product
                    # and attach it to the mapping or create one and attach
                    price_response = self.does_price_exist(product_id)
                    if price_response:
                        price_id = self.get_price_id_with_product(product_id)
                    else:
                        price_id = self.create_price_with_product(product_id)

                    self.products_mapping[product_name] = {
                        'product_id': product_id,
                        'price_id': price_id
                    }

    def sync_customers(self, site):
        logger.info(f"StripeProcessor sync_customers Started")
        stripe_customers = self.get_stripe_customers(site)
        stripe_customers_emails = [customer_obj['email'] for customer_obj in stripe_customers]

        vendor_customers_in_stripe = self.get_vendor_customers_in_stripe(stripe_customers_emails, site)

        vendor_customers_with_stripe_meta = vendor_customers_in_stripe.filter(meta__has_key='stripe_id')
        vendor_customers_without_stripe_meta = vendor_customers_in_stripe.exclude(meta__has_key='stripe_id')

        vendor_customers_not_in_stripe = self.get_vendor_customers_not_in_stripe(stripe_customers_emails, site)
        vendor_customer_to_create = vendor_customers_not_in_stripe | vendor_customers_without_stripe_meta

        self.create_stripe_customers(vendor_customer_to_create)
        self.update_stripe_customers(vendor_customers_with_stripe_meta)
        logger.info(f"StripeProcessor sync_customers Finished")

    def sync_offers(self, site):
        logger.info(f"StripeProcessor sync_offers Started")
        products = self.get_site_offers(site)
        offer_pk_list = [product['metadata']['pk'] for product in products]

        offers_in_vendor = self.get_vendor_offers_in_stripe(offer_pk_list, site)

        offers_in_vendor_with_stripe_meta = offers_in_vendor.filter(meta__has_key='stripe')
        offers_in_vendor_without_stripe_meta = offers_in_vendor.exclude(meta__has_key='stripe')
        
        offers_not_in_vendor = self.get_vendor_offers_not_in_stripe(offer_pk_list, site)
        offers_to_create = offers_not_in_vendor | offers_in_vendor_without_stripe_meta

        self.create_offers(offers_to_create)
        self.update_offers(offers_in_vendor_with_stripe_meta)
        logger.info(f"StripeProcessor sync_offers Finished")

    def sync_stripe_vendor_objects(self, site):
        """
        Sync up all the CustomerProfiles, Offers, Prices, and Coupons for all of the sites
        """
        logger.info(f"StripeProcessor sync_stripe_vendor_objects Started")
        self.sync_customers(site)
        self.sync_offers(site)
        logger.info(f"StripeProcessor sync_stripe_vendor_objects Finished")

    ##########
    # Stripe Transactions
    ##########
    def create_setup_intent(self, setup_intent_data):
        setup_intent = self.stripe_create_object(self.stripe.SetupIntent, setup_intent_data)

        return setup_intent

    def create_payment_intent(self, payment_intent_data):
        # Will return client secret value to be returned to the front end to continue processing payment

        #intent = self.stripe_call(self.stripe.PaymentIntent.create, {
        #    'customer': customer['id'],
        #    'setup_future_usage': 'off_session',
        #    'amount': self.invoice.get_one_time_transaction_total(),
        #    'currency': self.invoice.currency,
        #    'automatic_payment_methods': {
        #        'enabled': True
        #    }
        #})

        payment_intent = self.stripe_create_object(self.stripe.PaymentIntent, payment_intent_data)

        return payment_intent

    def set_stripe_payment_source(self):
        """
        This is needed for the charge api due to this error message:
        'You cannot create a charge with a PaymentMethod. Use the Payment Intents API instead'
        """
        if not self.source:
            if self.payment_info.is_valid():
                card_number = self.payment_info.cleaned_data.get('card_number')
                exp_month = self.payment_info.cleaned_data.get('expire_month')
                exp_year = self.payment_info.cleaned_data.get('expire_year')
                cvc = self.payment_info.cleaned_data.get('cvc_number')
                card = {
                    'number': card_number,
                    'exp_month': exp_month,
                    'exp_year': exp_year,
                    'cvc': cvc
                 }
                card = self.stripe_create_object(self.stripe.Token, {'card':card})
                if card:
                    self.source = card['id']

    def create_payment_method(self, payment_method_data):
        payment_method = self.stripe_create_object(self.stripe.PaymentMethod, payment_method_data)

        return payment_method

    def create_charge(self):
        charge_data = {
            'amount': self.to_stripe_valid_unit(self.invoice.get_one_time_transaction_total()),
            'currency': self.invoice.currency,
            'source': self.source,
        }
        charge = self.stripe_create_object(self.stripe.Charge, charge_data)
        if charge:
            return charge
        return None

    ##########
    # Base Processor Transaction Implementations
    ##########
    def pre_authorization(self):
        """
        Called before the authorization begins.
        """
        self.customer_setup()
        self.subscription_offer_setup()
        self.transaction_succeeded = False

    def process_payment(self):
        invoice_data = self.build_invoice()
        stripe_invoice = self.stripe_create_object(self.stripe.Invoice, invoice_data)
        if not stripe_invoice:
            return None

        payment_method_data = self.build_payment_method()
        stripe_payment_method = self.stripe_create_object(self.stripe.PaymentMethod, payment_method_data)
        if not stripe_payment_method:
            return None

        self.stripe_call(stripe_payment_method.attach, {'customer': self.invoice.profile.meta.get('stripe_id')})
        if not self.transaction_succeeded:
            return None
        
        self.invoice.vendor_notes['stripe_id'] = stripe_invoice.id
        self.invoice.save()
        
        stripe_line_items = []
        for order_item in self.invoice.get_one_time_transaction_order_items():
            line_item_data = self.build_invoice_line_item(order_item, stripe_invoice.id)
            stripe_line_item = self.stripe_create_object(self.stripe.InvoiceItem, line_item_data)
            stripe_line_items.append(stripe_line_item)
        
        stripe_invoice.lines = stripe_line_items

        amount = self.convert_decimal_to_integer(self.invoice.get_one_time_transaction_total())
        payment_intent_data = self.build_payment_intent(amount)
        stripe_payment_intent = self.stripe_create_object(self.stripe.PaymentIntent, payment_intent_data)

        if not stripe_payment_intent:
            return None

        self.stripe_call(stripe_invoice.pay, {"payment_method": stripe_payment_method.id})

        if self.transaction_succeeded:
            self.transaction_id = stripe_invoice.payment_intent

    def subscription_payment(self, subscription):
        payment_method_data = self.build_payment_method()
        stripe_payment_method = self.stripe_create_object(self.stripe.PaymentMethod, payment_method_data)
        if not stripe_payment_method:
            return None

        setup_intent_object = self.build_setup_intent(stripe_payment_method.id)
        if not setup_intent_object:
            return None

        stripe_setup_intent = self.stripe_create_object(self.stripe.SetupIntent, setup_intent_object)
        if not stripe_setup_intent:
            return None

        subscription_obj = self.build_subscription(subscription, stripe_payment_method.id)
        stripe_subscription = self.stripe_create_object(self.stripe.Subscription, subscription_obj)
        
        if not stripe_subscription or stripe_subscription.status == 'incomplete':
            self.transaction_succeeded = False
            return None

        if self.invoice.vendor_notes is None:
            self.invoice.vendor_notes = {}

        self.invoice.vendor_notes['stripe_id'] = stripe_subscription.latest_invoice
        self.invoice.save()

        self.subscription_id = stripe_subscription.id

    def charge_customer_profile(self):
        invoice_data = self.build_invoice()
        stripe_invoice = self.stripe_create_object(self.stripe.Invoice, invoice_data)
        if not stripe_invoice:
            return None

        stripe_customer_payment_methods = self.get_customer_payment_methods(self.invoice.profile.meta['stripe_id'])
        if not stripe_customer_payment_methods:
            return None
        
        self.invoice.vendor_notes['stripe_id'] = stripe_invoice.id
        self.invoice.save()
        
        stripe_line_items = []
        for order_item in self.invoice.get_one_time_transaction_order_items():
            line_item_data = self.build_invoice_line_item(order_item, stripe_invoice.id)
            stripe_line_item = self.stripe_create_object(self.stripe.InvoiceItem, line_item_data)
            stripe_line_items.append(stripe_line_item)
        
        stripe_invoice.lines = stripe_line_items

        amount = self.convert_decimal_to_integer(self.invoice.get_one_time_transaction_total())
        payment_intent_data = self.build_payment_intent(amount)
        stripe_payment_intent = self.stripe_create_object(self.stripe.PaymentIntent, payment_intent_data)

        if not stripe_payment_intent:
            return None

        self.stripe_call(stripe_invoice.pay, {"payment_method": stripe_customer_payment_methods[0].id})

        if self.transaction_succeeded:
            self.transaction_id = stripe_invoice.payment_intent

    def subscription_cancel(self, subscription):
        super().subscription_cancel(subscription)
        self.stripe_delete_object(self.stripe.Subscription, subscription.gateway_id)

    def subscription_update_payment(self, subscription):
        """
        Updates the credit card information for the subscription in stripe
        and updates the subscription model in vendor.
        """

        stripe_subscription_object = self.stripe_get_object(self.stripe.Subscription, subscription.gateway_id)
        if not stripe_subscription_object:
            return None

        current_payment_method = self.stripe.Customer.retrieve_payment_method(
            stripe_subscription_object.customer,
            stripe_subscription_object.default_payment_method
        )

        payment_method_data = {
            'type': 'card',
            'card': {
                'number': self.payment_info.cleaned_data.get('card_number'),
                'exp_month': self.payment_info.cleaned_data.get('expire_month'),
                'exp_year': self.payment_info.cleaned_data.get('expire_year'),
                'cvc': self.payment_info.cleaned_data.get('cvv_number'),
            },
            'billing_details': {
                'name': self.payment_info.cleaned_data.get('full_name', None),
                'email': subscription.profile.user.email
            }
        }
        # keep previous address
        if current_payment_method:
            payment_method_data['billing_details']['address'] = current_payment_method.billing_details.get("address")

        # create payment method using new card
        stripe_payment_method = self.stripe_create_object(self.stripe.PaymentMethod, payment_method_data)
        if not stripe_payment_method:
            return None

        setup_intent_object = {
            'customer': subscription.profile.meta['stripe_id'],
            'confirm': True,
            'payment_method_types': ['card'],
            'payment_method': stripe_payment_method.id,
            'metadata': {'site': subscription.profile.site}
        }

        # validate
        stripe_setup_intent = self.stripe_create_object(self.stripe.SetupIntent, setup_intent_object)
        if not stripe_setup_intent:
            return None

        # update subscription
        self.stripe.Subscription.modify(
            subscription.gateway_id,
            default_payment_method=stripe_payment_method
        )

        # save payment info to subscription model
        payment_info = {}
        account_number = payment_method_data.get("card", {}).get("number", "")[-4:]
        account_type = stripe_payment_method.card.get("brand", "")

        if account_number:
            payment_info['account_number'] = account_number

        if account_type:
            payment_info['account_type'] = account_type

        subscription.save_payment_info(payment_info)



import stripe

from core.models import Product
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from django.db.models import Q, signals
from django.test import Client, TestCase
from django.utils import timezone
from random import randrange
from siteconfigs.models import SiteConfigModel

from unittest import skipIf
from vendor.forms import BillingAddressForm, CreditCardForm
from vendor.models import CustomerProfile, Invoice, Offer, Price
from vendor.models.choice import InvoiceStatus
from vendor.processors import StripeProcessor

User = get_user_model()

@skipIf((settings.STRIPE_PUBLIC_KEY or settings.STRIPE_SECRET_KEY) is None, "Stripe enviornment variables not set, skipping tests")
class StripeProcessorTests(TestCase):
    fixtures = ['user', 'unit_test']

    VALID_CARD_NUMBERS = [
        '4242424242424242',  # visa
        '4000056655665556',  # visa debit
        '5555555555554444',  # mastercard
        '5200828282828210',  # mastercard debit
    ]

    def setup_processor_site_config(self):
        self.processor_site_config = SiteConfigModel()
        self.processor_site_config.site = self.existing_invoice.site
        self.processor_site_config.key = 'vendor.config.PaymentProcessorSiteConfig'
        self.processor_site_config.value = {"payment_processor": "stripe.StripeProcessor"}
        self.processor_site_config.save()

    def setup_user_client(self):
        self.client = Client()
        self.user = User.objects.get(pk=1)
        self.customer = CustomerProfile.objects.get(pk=1)
        self.client.force_login(self.user)

    def setup_existing_invoice(self):
        t_shirt = Product.objects.get(pk=1)
        t_shirt.meta['msrp']['usd'] = randrange(1, 1000)
        t_shirt.save()
        self.subscription_offer = Offer.objects.get(pk=6)
        price = Price.objects.get(pk=1)
        price.cost = randrange(1, 1000)
        price.priority = 10
        price.save()
        subscription_price = Price.objects.get(pk=9)
        subscription_price.cost = randrange(1, 1000)
        price.priority = 10
        subscription_price.priority = 10
        subscription_price.save()
        self.existing_invoice.update_totals()

    def setUp(self):
        self.setup_user_client()
        self.existing_invoice = Invoice.objects.get(pk=1)
        self.setup_processor_site_config()
        self.setup_existing_invoice()
        self.site = self.processor_site_config.site
        self.site.domain = 'sc'
        self.site.save()
        self.processor = StripeProcessor(self.site, self.existing_invoice)
        self.form_data = {
            'billing_address_form': {
                'billing-name': 'Home',
                'billing-company': 'Whitemoon Dreams',
                'billing-country': '840',
                'billing-address_1': '221B Baker Street',
                'billing-address_2': '',
                'billing-locality': 'Marylebone',
                'billing-state': 'California',
                'billing-postal_code': '90292'
            },
            'credit_card_form': {
                'full_name': 'Bob Ross',
                'card_number': '4242424242424242',
                'expire_month': '12',
                'expire_year': '2030',
                'cvv_number': '900',
                'payment_type': '10'
            }
        }
        self.valid_metadata = {'site': self.site.domain, 'pk': 1}
        self.valid_addr = {'city': "na", 'country': "US", 'line1': "Salvatierra walk", 'postal_code': "90321",
                           'state': 'CA'}

        self.cus_norrin_radd = {'name': 'Norrin Radd', 'email': 'norrin@radd.com', 'metadata': self.valid_metadata}
        self.cus_norrin_radd2 = {'name': 'Jake Paul', 'email': 'jpaul@radd.com', 'metadata': self.valid_metadata}
        self.pro_annual_license = {"name": "Annual Subscription", 'metadata': self.valid_metadata}
        self.pro_annual_license2 = {"name": "Annual Subscription 2", 'metadata': self.valid_metadata}
        self.pri_monthly = {"currency": "usd", "unit_amount": 1024,
                            "recurring": {"interval": "month", "interval_count": 1, "usage_type": "licensed"},
                            'metadata': self.valid_metadata}

    def test_environment_variables_set(self):
        self.assertIsNotNone(settings.STRIPE_PUBLIC_KEY)

    def test_processor_initialization_success(self):
        self.assertEquals(self.processor.provider, 'StripeProcessor')
        self.assertIsNotNone(self.processor.invoice)
        self.assertIsNotNone(self.processor.credentials)

    def test_process_payment_transaction_success(self):
        self.processor.create_stripe_customers([self.customer])
        self.processor.invoice.profile.refresh_from_db()
        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)
        self.processor.invoice.total = randrange(1, 1000)

        for recurring_order_items in self.processor.invoice.get_recurring_order_items():
            self.processor.invoice.remove_offer(recurring_order_items.offer)

        self.processor.set_stripe_payment_source()
        self.processor.transaction_succeeded = False

        self.processor.authorize_payment()

        self.assertIsNotNone(self.processor.payment)
        self.assertTrue(self.processor.payment.success)
        self.assertEquals(InvoiceStatus.COMPLETE, self.processor.invoice.status)

    def test_process_payment_transaction_fail_invalid_card(self):
        """
        Simulates a payment transaction for a reqeust.POST, with the payment and
        billing information. The test send an invalid card number to test the
        transation fails
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['card_number'] = '4242424242424241'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)
        self.processor.authorize_payment()

        self.assertIsNone(self.processor.payment)
        self.assertFalse(self.processor.transaction_succeeded)
        self.assertEquals(InvoiceStatus.CART, self.processor.invoice.status)

    def test_process_payment_transaction_fail_invalid_expiration(self):
        """
        Simulates a payment transaction for a reqeust.POST, with the payment and
        billing information. The test send an invalid expiration date to test the
        transation fails.
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['expire_month'] = str(timezone.now().month)
        self.form_data['credit_card_form']['expire_year'] = str(timezone.now().year - 1)

        self.processor.set_billing_address_form_data(self.form_data['billing_address_form'], BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data['credit_card_form'], CreditCardForm)

        self.processor.authorize_payment()

        self.assertIsNone(self.processor.payment)
        self.assertFalse(self.processor.transaction_succeeded)
        self.assertEquals(InvoiceStatus.CART, self.processor.invoice.status)

    def test_process_payment_fail_cvv_no_match(self):
        """
        Check incorrect cvc. Will fail with card number 4000000000000127
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '901'
        self.form_data['credit_card_form']['card_number'] = '4000000000000127'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_generic_decline(self):
        """
        Check a failed transaction due to to generic decline
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '902'
        self.form_data['credit_card_form']['card_number'] = '4000000000000002'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_fail_cvv_check_fails(self):
        """
        CVC number check fails for any cvv number passed
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '903'
        self.form_data['credit_card_form']['card_number'] = '4000000000000101'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_fail_expired_card(self):
        """
        Payment fails because of expired card
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '904'
        self.form_data['credit_card_form']['card_number'] = '4000000000000069'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_fail_fraud_always_blocked(self):
        """
        Fraud prevention fail: Always blocked
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '903'
        self.form_data['credit_card_form']['card_number'] = '4000000000000101'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_fail_fraud_higest_risk(self):
        """
        Fraud prevention fail: Higest Risk
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '903'
        self.form_data['credit_card_form']['card_number'] = '4000000000004954'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()

        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_fail_fraud_elevated_risk(self):
        """
        Fraud prevention fail : Elevated risk
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['cvv_number'] = '903'
        self.form_data['credit_card_form']['card_number'] = '4000000000009235'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()

        self.assertFalse(self.processor.transaction_succeeded)

    def test_process_payment_postal_code_check_fails(self):
        """
        Postal code check fails for any code given fo this card number
        """
        self.processor.create_stripe_customers([self.customer])
        self.processor.transaction_succeeded = False
        self.processor.invoice.profile.refresh_from_db()
        self.form_data['credit_card_form']['card_number'] = '4000000000000036'

        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)

        self.processor.invoice.total = randrange(1, 1000)
        self.processor.authorize_payment()
        self.assertFalse(self.processor.transaction_succeeded)

    def test_build_search_query_name(self):
        """
        Check our query string for stripe searches are valid
        """
        valid_query = 'name:"Johns Offer"'

        name_clause = self.processor.query_builder.make_clause_template(
            field='name',
            value='Johns Offer',
            operator=self.processor.query_builder.EXACT_MATCH
        )

        query = self.processor.query_builder.build_search_query(self.processor.stripe.Product, [name_clause])
        self.assertEquals(valid_query, query)

    def test_build_search_query_name_and_metadata(self):
        """
        Check our query string for stripe searches are valid
        """
        valid_query = 'name:"Johns Offer" AND metadata["site"]:"site4"'

        name_clause = self.processor.query_builder.make_clause_template(
            field='name',
            value='Johns Offer',
            operator=self.processor.query_builder.EXACT_MATCH,
            next_operator=self.processor.query_builder.AND
        )
        metadata_clause = self.processor.query_builder.make_clause_template(
            field='metadata',
            key='site',
            value='site4',
            operator=self.processor.query_builder.EXACT_MATCH
        )

        query = self.processor.query_builder.build_search_query(self.processor.stripe.Product, [name_clause, metadata_clause])
        self.assertEquals(valid_query, query)

    def test_build_search_query_metadata(self):
        """
        Check our query string for stripe searches are valid
        """
        valid_query = 'metadata["pk"]:4'

        metadata_clause = self.processor.query_builder.make_clause_template(
            field='metadata',
            key='pk',
            value=4,
            operator=self.processor.query_builder.EXACT_MATCH
        )

        query = self.processor.query_builder.build_search_query(self.processor.stripe.Product, [metadata_clause])
        self.assertEquals(valid_query, query)

    def test_build_search_query_metadata_fail(self):
        """
        Check metadata doesnt create query without key
        """

        metadata_clause = self.processor.query_builder.make_clause_template(
            field='metadata',
            value='site4',
            operator=self.processor.query_builder.EXACT_MATCH
        )

        query = self.processor.query_builder.build_search_query(self.processor.stripe.Product, [metadata_clause])
        self.assertEquals(query, "")

    def test_get_stripe_offers(self):
        # Test will fail on initial run without products created on stripe. Dont create duplicates

        offers = self.processor.get_site_offers(self.site)
        stripe_offer_names = [product['name'] for product in offers]
        names = [self.pro_annual_license['name'], self.pro_annual_license2['name']]

        offers_exist = [name for name in names if name in stripe_offer_names] or False

        if not offers_exist:
            self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)
            self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license2)

        current_stripe_offers = self.processor.get_site_offers(self.site)
        offer_names = [offer.name for offer in current_stripe_offers]

        self.assertIsNotNone(current_stripe_offers)
        self.assertIn(names[0], offer_names)
        self.assertIn(names[1], offer_names)

    def test_get_vendor_offers_in_stripe(self):
        # Test will fail on initial run without products created on stripe. Dont create duplicates

        offer1 = Offer.objects.create(site=self.site, name=self.pro_annual_license['name'], start_date=timezone.now())
        offer2 = Offer.objects.create(site=self.site, name=self.pro_annual_license2['name'], start_date=timezone.now())

        offers = self.processor.get_site_offers(self.site)
        stripe_offer_names = [product['name'] for product in offers]
        vendor_offer_names = [offer1.name, offer2.name]
        offers_exist = [name for name in vendor_offer_names if name in stripe_offer_names] or False

        if not offers_exist:
            self.pro_annual_license['metadata']['pk'] = offer1.pk
            self.pro_annual_license2['metadata']['pk'] = offer2.pk
            stripe_product1 = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)
            stripe_product2 = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license2)

        pk_list = [product['metadata']['pk'] for product in offers]
        vendor_offers_in_stripe = self.processor.get_vendor_offers_in_stripe(pk_list, self.site)

        self.assertIsNotNone(vendor_offers_in_stripe)
        self.assertIn(offer1, vendor_offers_in_stripe)
        self.assertIn(offer2, vendor_offers_in_stripe)

    def test_get_vendor_offers_not_in_stripe(self):
        offers = self.processor.get_site_offers(self.site)
        #stripe_offer_names = [product['name'] for product in offers]
        #vendor_offer_names = [self.pro_annual_license['name'], self.pro_annual_license2['name']]

        pk_list = [product['metadata']['pk'] for product in offers]

        vendor_offers_not_in_stripe = self.processor.get_vendor_offers_not_in_stripe(pk_list, self.site)

        self.assertIsNotNone(vendor_offers_not_in_stripe)

    def test_get_stripe_customers(self):
        # Test will fail on initial run without customers created on stripe. Dont create duplicates

        current_stripe_customers = self.processor.get_stripe_customers(self.site)
        customer_names = [customer.name for customer in current_stripe_customers]

        test_names = [self.cus_norrin_radd['name'], self.cus_norrin_radd2['name']]
        customers_exists = [name for name in test_names if name in customer_names] or False

        if not customers_exists:
            self.stripe_customer1 = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)
            self.stripe_customer2 = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd2)


        self.assertIsNotNone(current_stripe_customers)
        self.assertIn(self.cus_norrin_radd['name'], customer_names)
        self.assertIn(self.cus_norrin_radd2['name'], customer_names)

    def test_get_vendor_customers_in_stripe(self):
        # Test will fail on initial run without customers created on stripe. Dont create duplicates

        current_stripe_customers = self.processor.get_stripe_customers(self.site)
        customer_names = [customer.name for customer in current_stripe_customers]
        email_list = [customer.email for customer in current_stripe_customers]

        test_names = [self.cus_norrin_radd['name'], self.cus_norrin_radd2['name']]
        customers_exists = [name for name in test_names if name in customer_names] or False

        if not customers_exists:
            self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)
            self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd2)

        first_name1, last_name1 = self.cus_norrin_radd['name'].split(' ')
        first_name2, last_name2 = self.cus_norrin_radd2['name'].split(' ')
        user1 = User.objects.create(email=self.cus_norrin_radd['email'], first_name=first_name1, last_name=last_name1,
                                    username='test1')
        user2 = User.objects.create(email=self.cus_norrin_radd2['email'], first_name=first_name2, last_name=last_name2,
                                    username='test2')

        vendor_customer1 = CustomerProfile.objects.create(site=self.site, user=user1)
        vendor_customer2 = CustomerProfile.objects.create(site=self.site, user=user2)

        customers_in_vendor = self.processor.get_vendor_customers_in_stripe(email_list, self.site)
        customers_in_vendor_emails = [customer.user.email for customer in customers_in_vendor]

        self.assertIsNotNone(customers_in_vendor)
        self.assertIn(vendor_customer1.user.email, customers_in_vendor_emails)
        self.assertIn(vendor_customer2.user.email, customers_in_vendor_emails)

    def test_get_vendor_customers_not_in_stripe(self):
        current_stripe_customers = self.processor.get_stripe_customers(self.site)
        customer_names = [customer.name for customer in current_stripe_customers]
        email_list = [customer.email for customer in current_stripe_customers]

        test_names = [self.cus_norrin_radd['name'], self.cus_norrin_radd2['name']]
        customers_exists = [name for name in test_names if name in customer_names] or False
        if not customers_exists:
            self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)
            self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd2)

        customers_not_in_vendor = self.processor.get_vendor_customers_not_in_stripe(email_list, self.site)

        self.assertIsNotNone(customers_not_in_vendor)

    def test_create_stripe_customers(self):

        first_name1, last_name1 = self.cus_norrin_radd['name'].split(' ')
        first_name2, last_name2 = self.cus_norrin_radd2['name'].split(' ')
        user1 = User.objects.create(email=self.cus_norrin_radd['email'], first_name=first_name1, last_name=last_name1, username='test1')
        user2 = User.objects.create(email=self.cus_norrin_radd2['email'], first_name=first_name2, last_name=last_name2, username='test2')
        vendor_customer1 = CustomerProfile.objects.create(site=self.site, user=user1)
        vendor_customer2 = CustomerProfile.objects.create(site=self.site, user=user2)

        vendor_customers_not_in_stripe = [vendor_customer1, vendor_customer2]

        self.processor.create_stripe_customers(vendor_customers_not_in_stripe)

        vendor_customer1 = CustomerProfile.objects.get(site=self.site, user=user1)
        vendor_customer2 = CustomerProfile.objects.get(site=self.site, user=user2)

        self.processor.stripe_delete_object(self.processor.stripe.Customer, vendor_customer1.meta['stripe_id'])
        self.processor.stripe_delete_object(self.processor.stripe.Customer, vendor_customer2.meta['stripe_id'])

        self.assertIsNotNone(vendor_customer1.meta.get('stripe_id', None))
        self.assertIsNotNone(vendor_customer2.meta.get('stripe_id', None))

    def test_update_stripe_customers(self):
        stripe_customer1 = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)
        stripe_customer2 = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd2)

        user1 = User.objects.create(email=self.cus_norrin_radd['email'], first_name='New First Name1', last_name='Last Name1', username='test1')
        user2 = User.objects.create(email=self.cus_norrin_radd2['email'], first_name='New First Name2', last_name='Last Name2', username='test2')
        vendor_customer1 = CustomerProfile.objects.create(site=self.site, user=user1, meta={'stripe_id': stripe_customer1.id})
        vendor_customer2 = CustomerProfile.objects.create(site=self.site, user=user2, meta={'stripe_id': stripe_customer2.id})
        vendor_customers_in_stripe = [vendor_customer1, vendor_customer2]

        self.processor.update_stripe_customers(vendor_customers_in_stripe)

        updated_stripe_customer1 = self.processor.stripe_get_object(self.processor.stripe.Customer, stripe_customer1.id)
        updated_stripe_customer2 = self.processor.stripe_get_object(self.processor.stripe.Customer, stripe_customer2.id)

        self.processor.stripe_delete_object(self.processor.stripe.Customer, stripe_customer1.id)
        self.processor.stripe_delete_object(self.processor.stripe.Customer, stripe_customer2.id)

        self.assertEquals(updated_stripe_customer1.name, f'{user1.first_name} {user1.last_name}')
        self.assertEquals(updated_stripe_customer2.name, f'{user2.first_name} {user2.last_name}')

    def test_check_product_does_exist(self):
        offers = self.processor.get_site_offers(self.site)
        stripe_offer_names = [product['name'] for product in offers]

        offer_exists = self.pro_annual_license['name'] in stripe_offer_names

        if not offer_exists:
            self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)

        metadata = {
                'key': 'site',
                'value': 'site4',
                'field': 'metadata'
        }
        product = self.processor.does_product_exist(self.pro_annual_license['name'], metadata=metadata)

        self.assertIsNotNone(product)

    def test_get_product_id_with_name(self):
        # Test will fail on initial run without products created on stripe. Dont create duplicates

        metadata = {
            'key': 'site',
            'value': self.site.domain,
            'field': 'metadata'
        }
        product_id = self.processor.get_product_id_with_name(self.pro_annual_license['name'], metadata=metadata)

        offers = self.processor.get_site_offers(self.site)
        stripe_offer_ids = [product['id'] for product in offers]

        self.assertIn(product_id, stripe_offer_ids)

    def test_get_customer_payment_methods(self):
        payment_methods = self.processor.get_customer_payment_methods("cus_Mt98i0KskNWP0X")

        self.assertTrue(payment_methods)

    def test_charge_customer_profile_success(self):
        self.processor.invoice.profile.meta['stripe_id'] = "cus_Mt98i0KskNWP0X"
        self.processor.invoice.profile.save()
        self.processor.invoice.profile.refresh_from_db()
        self.processor.set_billing_address_form_data(self.form_data.get('billing_address_form'), BillingAddressForm)
        self.processor.set_payment_info_form_data(self.form_data.get('credit_card_form'), CreditCardForm)
        self.processor.invoice.total = randrange(1, 1000)

        for recurring_order_items in self.processor.invoice.get_recurring_order_items():
            self.processor.invoice.remove_offer(recurring_order_items.offer)

        self.processor.set_stripe_payment_source()
        self.processor.transaction_succeeded = False

        self.processor.process_customer_profile_payment()

        self.assertIsNotNone(self.processor.payment)
        self.assertTrue(self.processor.payment.success)
        self.assertEquals(InvoiceStatus.COMPLETE, self.processor.invoice.status)


    """
    Commenting out since stripe doesnt allow you to delete Price objects (weird)

    def test_check_price_does_exist(self):
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)
        self.pri_monthly['product'] = stripe_product.id
        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, self.pri_monthly)

        metadata = {
            'key': 'site',
            'value': 'site4',
            'field': 'metadata'
        }
        price = self.processor.does_price_exist(self.pri_monthly['product'], metadata=metadata)

        self.assertIsNotNone(price)


        self.processor.stripe_delete_object(self.processor.stripe.Product, stripe_product.id)
        self.processor.stripe_delete_object(self.processor.stripe.Price, stripe_price.id)"""

    def test_sync_customers(self):
        # Vendor objects not in stripe so create them there

        user = User.objects.get(pk=1)
        customer = CustomerProfile.objects.create(site=self.site, user=user)

        self.processor.sync_customers(customer.site)

        customer.refresh_from_db()

        stripe_id = customer.meta.get('stripe_id')

        self.processor.stripe_delete_object(self.processor.stripe.Customer, stripe_id)
        self.assertTrue(stripe_id)

    def test_sync_offers(self):
        signals.post_save.disconnect(receiver=signals.post_save, sender=CustomerProfile)
        now = timezone.now()

        offer = Offer.objects.create(site=self.site, name='Stripe Offer', start_date=now)
        product = Product.objects.create(name='Stripe Product', site=self.site)
        product.offers.add(offer)
        price = Price.objects.create(offer=offer, cost=10.99, start_date=timezone.now())

        # clear offers that dont have a price, since we cant create stripe product and price
        offers_with_no_price = Offer.objects.filter(Q(site=offer.site), Q(prices=None) | Q(prices__start_date__lte=now))
        offers_with_no_price.delete()

        self.processor.sync_offers(offer.site)

        offer.refresh_from_db()
        stripe_meta = offer.meta.get('stripe')


        self.assertTrue(stripe_meta)
        self.assertTrue(stripe_meta.get('product_id'))
        self.assertTrue(stripe_meta.get('price_id'))


@skipIf((settings.STRIPE_PUBLIC_KEY or settings.STRIPE_SECRET_KEY) is None, "Strip enviornment variables not set, skipping tests")
class StripeCRUDObjectTests(TestCase):

    def init_test_objects(self):
        self.valid_metadata = {'site': 'sc', 'pk':1}
        self.valid_addr = {'city': "na",'country': "US",'line1': "Salvatierra walk",'postal_code': "90321",'state': 'CA'}

        self.cus_norrin_radd = {'name': 'Norrin Radd', 'email': 'norrin@radd.com', 'metadata': self.valid_metadata}

        self.pro_monthly_license = {'name': "Monthly License", 'metadata': self.valid_metadata}
        self.pro_annual_license = {"name": "Annual Subscription", 'metadata': self.valid_metadata}

        self.pri_monthly = {"currency": "usd", "unit_amount": 1024, "recurring": {"interval": "month", "interval_count": 1, "usage_type": "licensed"}, 'metadata': self.valid_metadata}
        self.card = {'number': 4242424242424242, 'exp_month': "10", 'exp_year': "2023", 'cvc': "9000"}
        self.payment_method = {'type': 'card', 'card': self.card}
        self.cou_first_three_months_coupon = {
            "currency": "usd",
            "duration": "repeating",
            "duration_in_months": 3,
            "name": "25.5% off",
            "percent_off": 25.5,
            'metadata': self.valid_metadata
            }
        self.cou_first_month_free = {
            "currency": "usd",
            "duration": "once",
            "name": "100% off",
            "percent_off": 100,
            'metadata': self.valid_metadata
            }

    def setUp(self):
        stripe.api_key = settings.STRIPE_PUBLIC_KEY
        self.site = Site.objects.get(pk=1)
        self.site.domain = 'sc'
        self.site.save()
        self.init_test_objects()
        self.processor = StripeProcessor(self.site)

    ##########
    # Customer CRUD
    def test_create_customer_success(self):
        stripe_customer = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)

        self.processor.stripe_delete_object(self.processor.stripe.Customer, stripe_customer.id)

        self.assertIsNotNone(stripe_customer.id)


    def test_create_customer_with_address_success(self):
        self.cus_norrin_radd['address'] = self.valid_addr

        stripe_customer = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)

        self.processor.stripe_delete_object(self.processor.stripe.Customer, stripe_customer.id)

        self.assertIsNotNone(stripe_customer.id)


    ##########
    # Product CRUD
    def test_create_product_success(self):
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)
        self.processor.stripe_delete_object(self.processor.stripe.Product, stripe_product.id)

        self.assertIsNotNone(stripe_product.id)

    def test_create_product_no_name_fail(self):
        del(self.pro_monthly_license['name'])

        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_monthly_license)
        self.assertFalse(self.processor.transaction_succeeded)

    def test_get_product_success(self):
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)

        self.assertIsNotNone(stripe_product.id)

        fetch_product = self.processor.stripe_get_object(self.processor.stripe.Product, stripe_product.id)

        self.processor.stripe_delete_object(self.processor.stripe.Product, stripe_product.id)

        self.assertIsNotNone(fetch_product.id)

    def test_update_product_success(self):
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)

        self.assertIsNotNone(stripe_product.id)

        update_data = {"name": "Annual Subscription 2"}
        update_product = self.processor.stripe_update_object(self.processor.stripe.Product, stripe_product.id, update_data)

        self.processor.stripe_delete_object(self.processor.stripe.Product, stripe_product.id)

        self.assertIsNotNone(update_product.id)
        self.assertEquals(update_product['name'], update_data['name'])


    ##########
    # Price CRUD
    def test_create_price_product_data_success(self):
        del(self.pro_monthly_license['metadata'])
        self.pri_monthly['product_data'] = self.pro_monthly_license

        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, self.pri_monthly)

        self.assertIsNotNone(stripe_price.id)

    def test_create_price_product_id_success(self):
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_annual_license)
        self.pri_monthly['product'] = stripe_product.id

        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, self.pri_monthly)

        self.assertIsNotNone(stripe_price.id)

    def test_create_price_invalid_field_fail(self):
        self.pri_monthly['type'] = "This is not a valid field"

        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, self.pri_monthly)

        self.assertFalse(self.processor.transaction_succeeded)

    ##########
    # Coupon CRUD
    def test_create_coupon_success(self):
        stripe_coupon = self.processor.stripe_create_object(self.processor.stripe.Coupon, self.cou_first_month_free)

        stripe_coupon = self.processor.stripe_delete_object(self.processor.stripe.Coupon, stripe_coupon.id)

        self.assertIsNotNone(stripe_coupon.id)

    ##########
    # Subscription CRUD
    def test_create_subscription_success(self):
        stripe_cus_norrin_radd = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)

        stripe_payment_method = self.processor.stripe_create_object(self.processor.stripe.PaymentMethod, self.payment_method)

        setup_intent_object = {
            'customer': stripe_cus_norrin_radd.id,
            'confirm': True,
            'payment_method_types': ['card'],
            'payment_method': stripe_payment_method.id,
            'metadata': self.valid_metadata
        }
        stripe_setup_intent = self.processor.stripe_create_object(self.processor.stripe.SetupIntent, setup_intent_object)

        stripe_pro_monthly = self.processor.stripe_create_object(self.processor.stripe.Product, self.pro_monthly_license)

        self.pri_monthly['product'] = stripe_pro_monthly.id
        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, self.pri_monthly)

        subscription_obj = {
            'customer': stripe_cus_norrin_radd.id,
            'items': [{'price': stripe_price.id}],
            'default_payment_method': stripe_payment_method.id,
            'metadata': self.valid_metadata,

        }
        stripe_subscription = self.processor.stripe_create_object(self.processor.stripe.Subscription, subscription_obj)

        self.assertIsNotNone(stripe_subscription.id)

    ##########
    # Invoice CRUD
    # def test_create_invoice_success(self):

        # setup_intent_object = {
        #     'customer': stripe_cus_norrin_radd.id,
        #     'confirm': True,
        #     'payment_method_types': ['card'],
        #     'payment_method': stripe_payment_method.id,
        #     'metadata': self.valid_metadata
        # }


        # self.pri_monthly['product'] = stripe_pro_monthly.id

        # subscription_obj = {
        #     'customer': stripe_cus_norrin_radd.id,
        #     'items': [{'price': stripe_price.id}],
        #     'default_payment_method': stripe_payment_method.id,
        #     'metadata': self.valid_metadata,

        # }

        # stripe_invoice_object = {
        #     'metadata': self.valid_metadata,
        #     'subscription': stripe_sub.id,
        #     'customer': stripe_cus_norrin_radd.id,
        #     'statement_descriptor': 'Test description'
        # }


        # self.assertIsNotNone(stripe_invoice.id)

    ##########
    # Setup Intent CRUD
    def test_create_setup_intent_success(self):

        stripe_cus_norrin_radd = self.processor.stripe_create_object(self.processor.stripe.Customer, self.cus_norrin_radd)
        stripe_payment_method = self.processor.stripe_create_object(self.processor.stripe.PaymentMethod, self.payment_method)

        setup_intent_object = {
            'customer': stripe_cus_norrin_radd.id,
            'confirm': True,
            'payment_method_types': ['card'],
            'payment_method': stripe_payment_method.id,
            'metadata': self.valid_metadata
        }
        stripe_setup_intent = self.processor.stripe_create_object(self.processor.stripe.SetupIntent, setup_intent_object)

        self.assertIsNotNone(stripe_setup_intent.id)


@skipIf((settings.STRIPE_PUBLIC_KEY or settings.STRIPE_SECRET_KEY) is None, "Strip enviornment variables not set, skipping tests")
class StripeBuildObjectTests(TestCase):

    fixtures = ['user', 'unit_test']

    def setUp(self):
        stripe.api_key = settings.STRIPE_PUBLIC_KEY
        self.site = Site.objects.get(pk=1)
        self.site.domain = 'sc'
        self.site.save()
        self.processor = StripeProcessor(self.site)

    def test_build_customer_success(self):
        customer_profile = CustomerProfile.objects.all().first()

        customer_data = self.processor.build_customer(customer_profile)
        stripe_customer = self.processor.stripe_create_object(self.processor.stripe.Customer, customer_data)

        self.assertIsNotNone(stripe_customer.id)
        self.assertEqual(f"{customer_profile.user.first_name} {customer_profile.user.last_name}", stripe_customer.name)
        self.assertEqual(customer_profile.user.email, stripe_customer.email)

    def test_build_product_success(self):
        offer = Offer.objects.all().first()

        product_data = self.processor.build_product(offer)
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, product_data)

        self.assertIsNotNone(stripe_product.id)
        self.assertEqual(offer.name, stripe_product.name)

    def test_build_price_success(self):
        offer = Offer.objects.all().first()

        product_data = self.processor.build_product(offer)
        stripe_product = self.processor.stripe_create_object(self.processor.stripe.Product, product_data)

        offer.meta['stripe'] = {
            'product_id': stripe_product.id
        }
        offer.save()
        price = offer.prices.first()

        price_data = self.processor.build_price(offer, price)
        stripe_price = self.processor.stripe_create_object(self.processor.stripe.Price, price_data)

        self.assertIsNotNone(stripe_price.id)
        self.assertEqual(price.cost, stripe_price.unit_amount)

    def test_build_coupon_success(self):
        offer = Offer.objects.all().first()
        price = offer.prices.first()

        coupon_data = self.processor.build_coupon(offer, price)
        stripe_coupon = self.processor.stripe_create_object(self.processor.stripe.Coupon, coupon_data)

        self.assertIsNotNone(stripe_coupon.id)
        self.assertEqual("".join([str(stripe_coupon.amount_off)[:-2], ".", str(stripe_coupon.amount_off)[-2:]]), str((offer.get_msrp() - price.cost)))

    def test_build_subscription_success(self):
        pass

    def test_build_payment_method_successs(self):
        pass

    def test_build_setup_intent_success(self):
        pass

